"""
Avro file reader to pick out specific alert packet from combined avro file
"""
from fastavro import reader


def pick_alert(
    avro_path: str,
    candid: int,
) -> list:
    """
    Pick out specific alert packet from combined avro file

    Parameters:
    ----------
    avro_path : str
        Path to avro file
    candid: int
        Candidate ID; unique identifier

    Returns:
    ----------
    candidate_record: list
        alert packet
    """
    candidate_record = []
    with open(avro_path, "rb") as avro_f:
        avro_reader = reader(avro_f)
        for record in avro_reader:
            if record["candid"] == candid:
                candidate_record.append(record)

    return candidate_record



"""
Fundamental classification types for DRB.
"""

RB_CLASS = {
    -1: "Unclassified",
    0: "Unknown",
    1: "Real",
    2: "Bogus - Other",
    3: "Bogus - YinYang",
    4: "Bogus - Diffraction Spike",
    5: "Bogus - Bad Astrometry",
    6: "Bogus - Stripes",
    7: "Missing cutouts",
    8: "Bogus - Background",
    9: "Bogus - Failed imsub",
    10: "Bogus - Low S/N",
    11: "Bogus - Masking",
    12: "Bogus - Artifact",
    13: "Bogus - Ambiguous",
    14: "Real - Ambiguous",
}



import gzip
import io

import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from astropy.stats import sigma_clipped_stats
from fastavro import reader, writer
from scipy import ndimage


def make_triplet(alert, normalize: bool = False) -> np.ndarray:
    """
    Make stacked triplet from alert packet

    Parameters:
    ----------
    alert : dict
        Alert packet
    normalize: bool
        Normalize the cutout counts; False by default

    Returns:
    ----------
    triplet: np.ndarray
        Stacked triplet of science, reference and difference cutouts
    """
    cutout_dict = dict()

    for cutout in ["science", "template", "difference"]:
        cutout_zipped = alert["cutout_" + cutout]

        with gzip.open(io.BytesIO(cutout_zipped), "rb") as f:
            with fits.open(io.BytesIO(f.read())) as hdu:
                cut_data = hdu[0].data
                cutout_dict[cutout] = np.nan_to_num(cut_data)
                if normalize:
                    cutout_dict[cutout] /= np.linalg.norm(cutout_dict[cutout])
                    cutout_dict[cutout] -= np.nanmean(cutout_dict[cutout])

    shape = np.shape(cutout_dict[cutout]) + (3,)
    triplet = np.zeros(shape)
    triplet[:, :, 0] = cutout_dict["science"]
    triplet[:, :, 1] = cutout_dict["template"]
    triplet[:, :, 2] = cutout_dict["difference"]

    return triplet


def pick_alert(avro_path: str, candid: int, write_alert: bool = False) -> list:
    """
    Pick out specific alert packet from combined avro file

    Parameters:
    ----------
    avro_path : str
        Path to avro file
    candid: int
        Candidate ID; unique identifier

    Returns:
    ----------
    candidate_record: list
        alert packet
    """
    candidate_record = []
    with open(avro_path, "rb") as avro_f:
        avro_reader = reader(avro_f)
        schema = avro_reader.writer_schema
        for record in avro_reader:
            if record["candid"] == candid:
                candidate_record.append(record)
                if write_alert:
                    with open(str(candid) + ".avro", "wb") as out:
                        writer(out, schema, candidate_record)

    return candidate_record


def rotate_triplet(triplet):
    """
    Rotate the triplet by 90, 180 and 270 degrees

    Args:
        triplet: Triplet of cutouts

    Returns:

    """
    sci, ref, diff = triplet[:, :, 0], triplet[:, :, 1], triplet[:, :, 2]
    trip90 = np.stack(
        (ndimage.rotate(sci, 90), ndimage.rotate(ref, 90), ndimage.rotate(diff, 90)),
        axis=-1,
    )
    trip180 = np.stack(
        (ndimage.rotate(sci, 180), ndimage.rotate(ref, 180), ndimage.rotate(diff, 180)),
        axis=-1,
    )
    trip270 = np.stack(
        (ndimage.rotate(sci, 270), ndimage.rotate(ref, 270), ndimage.rotate(diff, 270)),
        axis=-1,
    )
    return trip90, trip180, trip270


def plot_triplet(stack: np.ndarray, pdf: bool = False) -> None:
    """
    Plot cutouts of sci, ref and diff from stacked triplet

    Parameters:
    ----------
    stack: np.ndarray
        Stack of cutouts
    pdf: bool
        Save plots in a pdf file

    Returns:
    ----------
    None
    """

    # read cutouts
    sci, ref, diff = stack[:, :, 0], stack[:, :, 1], stack[:, :, 2]

    # plot cutouts
    fig, ax = plt.subplots(1, 3, figsize=(12, 4))
    _, median, std = sigma_clipped_stats(sci)
    ax[0].imshow(sci, vmin=median - 5 * std, vmax=median + 5 * std, origin="lower")
    ax[0].set_title("sci")
    ax[0].axis("off")
    _, median, std = sigma_clipped_stats(ref)
    ax[1].imshow(ref, vmin=median - 5 * std, vmax=median + 5 * std, origin="lower")
    ax[1].set_title("ref")
    ax[1].axis("off")
    _, median, std = sigma_clipped_stats(diff)
    ax[2].imshow(diff, vmin=median - 5 * std, vmax=median + 5 * std, origin="lower")
    ax[2].set_title("diff")
    ax[2].axis("off")

    # save to pdf if necessary
    if pdf:
        pdf.savefig()
        plt.close()
    else:
        plt.show()



"""
Module for the model
"""

import torch
import torch.nn.functional as F
from torch import nn


class WINTERNet(nn.Module):
    """
    WINTERNet model
    """

    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=16, kernel_size=3, padding=1)
        # self.conv1 = nn.Conv2d(81, 16, kernel_size=3, stride=1, padding=1)
        self.conv2 = nn.Conv2d(
            in_channels=16, out_channels=16, kernel_size=3, padding=1
        )
        self.conv3 = nn.Conv2d(
            in_channels=16, out_channels=16, kernel_size=3, padding=1
        )
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.batch_norm1 = nn.BatchNorm2d(16, momentum=0.99, eps=1e-3)
        self.dropout1 = nn.Dropout2d(p=0.25)
        self.conv4 = nn.Conv2d(
            in_channels=16, out_channels=32, kernel_size=3, padding=1
        )
        self.conv5 = nn.Conv2d(
            in_channels=32, out_channels=32, kernel_size=3, padding=1
        )
        self.conv6 = nn.Conv2d(
            in_channels=32, out_channels=32, kernel_size=3, padding=1
        )
        self.pool2 = nn.MaxPool2d(kernel_size=4, stride=4)
        self.batch_norm2 = nn.BatchNorm2d(32, momentum=0.99, eps=1e-3)
        self.dropout2 = nn.Dropout2d(p=0.25)
        # self.fc1 = nn.Linear(32 * 5 * 5, 256)  # Calculated based on the size after applying max pooling
        self.fc1 = nn.Linear(3200, 256)
        self.batch_norm3 = nn.BatchNorm1d(256, momentum=0.99, eps=1e-3)
        self.dropout3 = nn.Dropout(p=0.5)
        self.fc_out = nn.Linear(256, 1)

    def forward(self, x):
        x = x.to(self.conv1.bias.dtype)
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = self.pool(x)
        x = self.batch_norm1(x)
        x = self.dropout1(x)
        x = F.relu(self.conv4(x))
        x = F.relu(self.conv5(x))
        x = F.relu(self.conv6(x))
        x = self.pool2(x)
        x = self.batch_norm2(x)
        x = self.dropout2(x)
        x = x.reshape(x.size(0), -1)
        x = F.relu(self.fc1(x))
        x = self.batch_norm3(x)
        x = self.dropout3(x)
        x = self.fc_out(x)
        x = torch.sigmoid(x)
        return x.view(-1, 1)



"""
Module for handling paths in the WinterRB project.
"""
import os
from pathlib import Path

data_dir = os.getenv("WINTERRB_DATA_DIR", None)
if data_dir is None:
    data_dir = Path.home().joinpath("Data/winter_real_bogus")

train_class_path = data_dir / "data_train.csv"

code_dir = Path(__file__).resolve().parent

model_dir = code_dir.parent / "models"
model_dir.mkdir(exist_ok=True)

